/*
   USB connection library
   by paul69, v0.02, 2003-01-23, <e2500@narod.ru>

   Supported camera:
      E995 (tested)
      E2500 (tested)
      E4300 (tested)
      E4500 (tested)
*/
#define STRICT
#include <windows.h>
#pragma hdrstop

#include <stdio.h>
#include "dsc_connect.h"

FILE* listing;

class USB_DSC : public DSC
{
public:

   USB_DSC(void* _handle, const char* _comment) : DSC(_handle, _comment)
   {
      if( *(long*)&comment[0] == *(long*)"\\\\.\\" )
         *(long*)&comment[0] = *(long*)"USB,";
   }
   
  ~USB_DSC();

   virtual bool SetReg(int reg, unsigned value);
   virtual bool GetReg(int reg, unsigned& value);

   virtual bool SetVar(int reg, const void* data, unsigned size);
   virtual bool SetStr(int reg, const char* str);
   virtual bool GetVar(int reg, TResponce&);

   virtual bool SpecCmd(int reg, unsigned value, unsigned timeout=0);

   virtual bool GetDeviceInfo(TDeviceInfo&);
   virtual bool CloseCamera();
};


////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//

struct TCmd
{
   dword command;
   dword size;
};

struct TCmdUSB
{
   dword command;
   dword _zero;

   byte size_hi;
   byte size_lo;
};

struct TScsiPassThrough
{
    word Length;
    byte ScsiStatus;
    byte PathId;
    byte TargetId;
    byte Lun;
    byte CdbLength;
    byte SenseInfoLength;
    byte DataIn;
    byte _align[3];
    dword DataTransferLength;
    dword TimeOutValue;
    void* DataBuffer;
    dword SenseInfoOffset;

    union
    {
       byte Cdb[16];
       
       // sizeof *this == 128 bytes
       byte buffer[128-7*4];

       TCmd Cmd;
       TCmdUSB USB;
    };
};

struct TScsiCommand
{
   byte command;
   byte data_in;
   word timeout;
};

static TScsiCommand command_table[] =
{
   { 0x12, 1, 12 },
   { 0xE0, 0, 2 },
   { 0xE1, 0, 2 },
   { 0xE2, 1, 50 },
   { 0xE3, 1, 2 },
   { 0xE4, 1, 50 },
   { 0xE5, 0, 2 },
};

int timeout;
const int n_commands = sizeof(command_table) / sizeof(command_table[0]);

enum USB_COMMAND
{
   READ_DEVICE_NAME, //12
   USB_START_IO,     //E0
   USB_SEND_CMD,     //E1
   USB_RCV_RESPONCE, //E2
   USB_RCV_ACK,      //E3
   USB_RCV_SIZE,     //E4
   USB_CLOSE,        //E5
};

#ifndef IOCTL_SCSI_PASS_THROUGH_DIRECT
#define IOCTL_SCSI_PASS_THROUGH_DIRECT 0x0004D014
#endif

bool SendCmd_to_DSC(HANDLE hDSC, USB_COMMAND i_command, void* buf, unsigned size)
{
   TScsiCommand command = command_table[i_command];

   TScsiPassThrough data;
   memset(&data, 0, sizeof(data));

   data.Length = 44;
   //data.ScsiStatus = 0;
   //data.PathId = 0;
   data.TargetId = 1;
   //data.Lun = 0;
   data.CdbLength = 16;
   data.SenseInfoLength = 24;
   data.DataIn = command.data_in;
   data.DataTransferLength = size;
   data.TimeOutValue = timeout ? timeout : command.timeout;
   data.DataBuffer = buf;
   data.SenseInfoOffset = 48;

   if(data.DataIn)
      memset(buf, 0, size);

   if( command.command == 0x12 )
   {
      data.CdbLength = 12;
      data.Cmd.command = command.command;
      data.Cmd.size = size;
   }
   else
   {
      data.USB.command = command.command;
      data.USB.size_lo = (byte)(size);
      data.USB.size_hi = (byte)(size >> 8);
   }

   DWORD BytesReturned;
   return 0 != DeviceIoControl( hDSC, IOCTL_SCSI_PASS_THROUGH_DIRECT,
      &data, sizeof(data), &data, sizeof(data), &BytesReturned, NULL );
}

struct TCmdAck
{
   dword length;
   word code; // 1,2,3
   word code2; // 0,0x9FFF
   dword _unknown[14]; // for sizeof == 64
   byte ack;
};

struct TInquiryData
{
    byte DeviceType : 5;
    byte DeviceTypeQualifier : 3;
    byte DeviceTypeModifier : 7;
    byte RemovableMedia : 1;
    byte Versions;
    byte ResponseDataFormat : 4;
    byte HiSupport : 1;
    byte NormACA : 1;
    byte ReservedBit : 1;
    byte AERC : 1;
    byte AdditionalLength;
    byte Reserved[2];
    byte SoftReset : 1;
    byte CommandQueue : 1;
    byte Reserved2 : 1;
    byte LinkedCommands : 1;
    byte Synchronous : 1;
    byte Wide16Bit : 1;
    byte Wide32Bit : 1;
    byte RelativeAddressing : 1;
    byte VendorId[8];
    byte ProductId[16];
    byte ProductRevisionLevel[4];
};

bool GetDeviceName(HANDLE hDSC, TInquiryData& inquiry)
{
   return SendCmd_to_DSC(hDSC, READ_DEVICE_NAME, &inquiry, sizeof(inquiry));
}

bool USB_DSC::GetDeviceInfo(TDeviceInfo& info)
{
   memset( &info, 0, sizeof(info) );

   TInquiryData inquiry;
   if( !GetDeviceName( GetHandle(), inquiry) )
      return false;

   memcpy( info.VendorId, inquiry.VendorId, sizeof(inquiry.VendorId) );
   info.VendorId[sizeof(inquiry.VendorId)] = '\0';

   memcpy( info.ProductId, inquiry.ProductId, sizeof(inquiry.ProductId) );
   info.ProductId[sizeof(inquiry.ProductId)] = '\0';

   memcpy( info.ProductRevisionLevel, inquiry.ProductRevisionLevel, sizeof(inquiry.ProductRevisionLevel) );
   info.ProductRevisionLevel[sizeof(inquiry.ProductRevisionLevel)] = '\0';
   return true;
}

static bool CloseCamera(HANDLE hDSC)
{
   return SendCmd_to_DSC(hDSC, USB_CLOSE, NULL, 0);
}

bool USB_DSC::CloseCamera()
{
   return ::CloseCamera( GetHandle() );
}

struct TSendStartIO
{
   dword length;
   word code; // 1
   word code2; // 0x9FFF
   dword _unknown[2]; // for sizeof == 16
};

static TSendStartIO _Start_IO = { 16, 1, 0x9FFF };

static bool usbStartIO(HANDLE hDSC)
{
   return SendCmd_to_DSC(hDSC, USB_START_IO, &_Start_IO, sizeof(_Start_IO) );
}

static bool usbSendCmd(HANDLE hDSC, int type, int reg)
{
   struct TSendCmd
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      word check_sum;
   } send;

   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 2;
   send.command = (byte)type;
   send.reg = (byte)reg;
   send.check_sum = (word)((unsigned)send.command + (unsigned)send.reg);

   return SendCmd_to_DSC(hDSC, USB_SEND_CMD, &send, send.length);
}

static bool usbSendCmdSetStr(HANDLE hDSC, int reg, const char* str)
{
   struct TSendCmdSetStr
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      byte data[1];
   };

   int len = strlen(str);
   int size = sizeof(TSendCmdSetStr)-1+len+2;

   TSendCmdSetStr& send = *(TSendCmdSetStr*)new byte[size];
   memset(&send, 0, size);

   send.length = size;
   send.code = 2;

   send.magic = 0x431B;
   send.size = (word)(len+2);
   send.command = 3;
   send.reg = (byte)reg;

   unsigned sum = (unsigned)send.command + (unsigned)send.reg;
   for(int i=0; i<len; ++i)
   {
      send.data[i] = (byte)str[i];
      sum += (unsigned)send.data[i];
   }
   send.data[len] = (byte)(sum);
   send.data[len+1] = (byte)(sum>>8);

   bool ret = SendCmd_to_DSC(hDSC, USB_SEND_CMD, &send, send.length);
   delete [] (byte*)&send;
   return ret;
}

static bool usbSendCmdWrite(HANDLE hDSC, int reg, const void* data, unsigned sz)
{
   struct TSendCmdWrite
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      byte data[2]; // two bytes checksum
   };

   int size = sizeof(TSendCmdWrite)+sz;

   TSendCmdWrite& send = *(TSendCmdWrite*)new byte[size];
   memset(&send, 0, size);

   send.length = size;
   send.code = 2;

   send.magic = (word)0x431B;
   send.size = (word)(sz+2);
   send.command = 3;
   send.reg = (byte)reg;

   unsigned sum = (unsigned)send.command + (unsigned)send.reg;
   for(int i=0; i<sz; ++i)
   {
      send.data[i] = ((byte*)data)[i];
      sum += (unsigned)send.data[i];
   }
   send.data[sz] = (byte)(sum);
   send.data[sz+1] = (byte)(sum>>8);

   bool ret = SendCmd_to_DSC(hDSC, USB_SEND_CMD, &send, send.length);
   delete [] (byte*)&send;
   return ret;
}

static bool usbSendCmdWrite2(HANDLE hDSC, int cmd, const void* data, unsigned sz)
{
   struct TSendCmdWrite2
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word cmd;
      word size;
      byte data[2]; // two bytes checksum
   };

   int size = sizeof(TSendCmdWrite2)+sz;

   TSendCmdWrite2& send = *(TSendCmdWrite2*)new byte[size];
   memset(&send, 0, size);

   send.length = size;
   send.code = 2;

   send.cmd = (word)cmd;
   send.size = (word)(sz+2);

   unsigned sum = 0;
   for(int i=0; i<sz; ++i)
   {
      send.data[i] = ((byte*)data)[i];
      sum += (unsigned)send.data[i];
   }
   send.data[sz] = (byte)(sum);
   send.data[sz+1] = (byte)(sum>>8);

   bool ret = SendCmd_to_DSC(hDSC, USB_SEND_CMD, &send, send.length);
   delete [] (byte*)&send;
   return ret;
}

static bool usbSendCmdSetReg(HANDLE hDSC, int reg, unsigned value)
{
   struct TSendCmdVar
   {
      dword length;
      word code; // 2
      word code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      dword value;
      word check_sum;
   } send;

   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 6;
   //send.command = (byte)0;
   send.reg = (byte)reg;
   send.value = value;
   send.check_sum = (word)
    ( (unsigned)(value&0xFF)
    + (unsigned)((value>>8)&0xFF)
    + (unsigned)((value>>16)&0xFF)
    + (unsigned)((value>>24)&0xFF)
    + (unsigned)send.reg);

   return SendCmd_to_DSC(hDSC, USB_SEND_CMD, &send, send.length);
}

static bool usbSpecCmd(HANDLE hDSC, int reg, int value)
{
   struct TSendCmd
   {
      dword length;
      word code; // 2
      word code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      byte value;
      word check_sum;
   } send;

   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 3;
   send.command = 2;
   send.reg = (byte)reg;
   send.value = (byte)value;
   send.check_sum = (word)(
      (unsigned)send.command +
      (unsigned)send.reg +
      (unsigned)send.value);

   return SendCmd_to_DSC(hDSC, USB_SEND_CMD, &send, send.length);
}

static bool usbSendAck(HANDLE hDSC)
{
   TCmdAck send;
   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;
   send.ack = 6;

   return SendCmd_to_DSC(hDSC, USB_SEND_CMD, &send, send.length);
}

static bool usbAcceptResponce(HANDLE hDSC, void* buf, int size)
{
   return SendCmd_to_DSC(hDSC, USB_RCV_RESPONCE, buf, size);
}

static bool usbGetAck(HANDLE hDSC)
{
   struct TResponceAck
   {
      dword length;
      dword code;
      word _unknown[3]; // for sizeof == 14
   } responce;

   if( !SendCmd_to_DSC(hDSC, USB_RCV_ACK, &responce, sizeof(responce)) ) return false;

   return (responce.code == 0x9FFF0003);
}

static bool usbGetSize(HANDLE hDSC, unsigned& size)
{
   struct TResponceSize
   {
      dword length;
      dword code;
      dword _unknown;
      dword responce_size;
   } responce;

   if( SendCmd_to_DSC(hDSC, USB_RCV_SIZE, &responce, sizeof(responce) ) )
   {
      size = responce.responce_size;
      return true;
   }
   return false;
}



//#define CHECK(x) if( !(x) ) { return false; }
#define CHECK(x) { if( !(x) ) { if(listing)fprintf(listing," %s:%d %s - fault\n", __FILE__, __LINE__, #x); return false; }}
//#define CHECK(x) { time(); if(listing)fprintf(listing,"%d: %s", __LINE__, #x); if( !(x) ) { elaps(); if(listing)fprintf(listing," - fault\n"); return false; } elaps(); if(listing)fprintf(listing,"\n"); }
#define CHECK2(x) { if(listing)fprintf(listing,"%d: %s\n", __LINE__, #x); if( !(x) ) { if(listing)fprintf(listing," - fault\n"); return false; } }

unsigned start_time;
unsigned prev_time;

void time()
{
   unsigned cur_time = GetTickCount();
   if( start_time == 0 )
      start_time = cur_time;

   if(listing)fprintf(listing,"%8d: ", cur_time - start_time);
   prev_time = cur_time;
}
void elaps()
{
   unsigned cur_time = GetTickCount();
   if(listing)fprintf(listing," (%d ms) ", cur_time - prev_time);
}

bool USB_DSC::GetVar(int reg, TResponce& responce)
{
   responce.free();

   struct TResponceVar
   {
      dword length;
      word code; // 1,2,3
      word code2; // 0,0x9FFF
      dword _unknown[14]; // for sizeof == 64

      byte cmd; // 3,2
      byte frame; // 0,...
      word len; //
      byte data[PACKET_SIZE+8];
   } resp;

   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbSendCmd(GetHandle(), 4, reg) );
   CHECK( usbGetAck(GetHandle()) );

   for(;;)
   {
      unsigned size;
      CHECK( usbStartIO(GetHandle()) );
      CHECK( usbGetSize(GetHandle(), size) );
      CHECK( size != 0 );
      CHECK( size <= sizeof(resp) );
      CHECK( usbAcceptResponce(GetHandle(), &resp, size));
      CHECK( usbGetAck(GetHandle()) );

      CHECK( usbStartIO(GetHandle()) );
      CHECK( usbSendAck(GetHandle()) );
      CHECK( usbGetAck(GetHandle()) );

      if(listing)fprintf(listing, "USB_DSC::GetVar cmd=%u frame=%u len=%u\n", resp.cmd, resp.frame, resp.len);

      if( resp.cmd == 3 )
      {
         responce.append(resp.data, resp.len);
         break;
      }

      CHECK( resp.cmd == 2 );

      responce.append(resp.data, resp.len);
   }
   return true;
}

bool USB_DSC::SetStr(int reg, const char* str)
{
   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbSendCmdSetStr(GetHandle(), reg, str) );
   CHECK( usbGetAck(GetHandle()) );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbGetSize(GetHandle(), size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(GetHandle(), &responce, size) );
   CHECK( usbGetAck(GetHandle()) );

   CHECK( responce.ack == 6 );

   return true;
}

bool USB_DSC::SetVar(int reg, const void* data, unsigned size)
{
   for(int i=0; size; ++i )
   {
      int cmd = (((i-1)&255)<<8) + 3;
      int sz = size;
      if( sz > PACKET_SIZE )
      {
         sz = PACKET_SIZE;
         --cmd;
      }

      CHECK( usbStartIO(GetHandle()) );
      CHECK( i == 0 ? usbSendCmdWrite(GetHandle(), reg, data, sz) : usbSendCmdWrite2(GetHandle(), cmd, data, sz) );
      CHECK( usbGetAck(GetHandle()) );

      unsigned size;
      TCmdAck responce;

      CHECK( usbStartIO(GetHandle()) );
      CHECK( usbGetSize(GetHandle(), size) );
      CHECK( size == sizeof(responce) );
      CHECK( usbAcceptResponce(GetHandle(), &responce, size) );
      CHECK( usbGetAck(GetHandle()) );

      CHECK( responce.ack == 6 );

      if( (size -= sz) == 0 ) break;
      data = (const void*)((unsigned)data + sz);
   }
   return true;
}

bool USB_DSC::GetReg(int reg, unsigned& value)
{
   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbSendCmd(GetHandle(), 1, reg) );
   CHECK( usbGetAck(GetHandle()) );

   value = -1;
   unsigned size;

   struct TResponceCmdGetReg
   {
      dword length;
      word code; // 1,2,3
      word code2; // 0,0x9FFF
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      dword value;
      word check_sum;
   } responce;

   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbGetSize(GetHandle(), size) );
   CHECK( size <= sizeof(responce) );
   CHECK( usbAcceptResponce(GetHandle(), &responce, size) );
   CHECK( usbGetAck(GetHandle()) );

   if( size == sizeof(TCmdAck) )
      if(listing)fprintf(listing,"responce=%02X\n", ((TCmdAck*)&responce)->ack);

   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbSendAck(GetHandle()) );
   CHECK( usbGetAck(GetHandle()) );

   CHECK( size == sizeof(responce) );

   value = responce.value;

   return true;
}

bool USB_DSC::SpecCmd(int reg, unsigned value, unsigned timeout)
{
   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbSpecCmd(GetHandle(), reg, value) );
   CHECK( usbGetAck(GetHandle()) );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbGetSize(GetHandle(), size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(GetHandle(), &responce, size) );
   CHECK( usbGetAck(GetHandle()) );

   CHECK( responce.ack == 6 );

   if( timeout )
      Sleep( timeout );

   return true;
}

bool USB_DSC::SetReg(int reg, unsigned value)
{
   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbSendCmdSetReg(GetHandle(), reg, value) );
   CHECK( usbGetAck(GetHandle()) );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO(GetHandle()) );
   CHECK( usbGetSize(GetHandle(), size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(GetHandle(), &responce, size) );
   CHECK( usbGetAck(GetHandle()) );

   if( responce.ack == 6 ) return true;

   if(listing)fprintf(listing,"responce=%02X\n", responce.ack);
   return false;
}

USB_DSC::~USB_DSC()
{
   if( GetHandle() == INVALID_HANDLE_VALUE )
      return;

   ::CloseCamera( GetHandle() );
}

static char DSC_DRIVE[] = "\\\\.\\C:";
static char DSC_PATH[] = "C:\\";

DSC* OpenUSB(char camera_drive)
{
   if( camera_drive != ' ' )
   {
      DSC_DRIVE[4] = camera_drive;

      HANDLE hDSC = CreateFile(DSC_DRIVE, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
         NULL, OPEN_EXISTING, 0, NULL );

      if( INVALID_HANDLE_VALUE == hDSC )
         return NULL;

      return new USB_DSC( hDSC, DSC_DRIVE );
   }

   DWORD drives = GetLogicalDrives();

   for(int i=2; i < 26; ++i)
   {
      if( drives & (1<<i) )
      {
         DSC_DRIVE[4] =
         DSC_PATH[0] = (char)('A' + i);

         if( GetDriveType( DSC_PATH ) != DRIVE_REMOVABLE	) continue;

         HANDLE hDSC = CreateFile(DSC_DRIVE, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
            NULL, OPEN_EXISTING, 0, NULL );

         if( hDSC == INVALID_HANDLE_VALUE ) continue;

         TInquiryData inquiry;
         bool ret = GetDeviceName(hDSC, inquiry);
         CloseCamera(hDSC);

         if( ret
            && memcmp(inquiry.VendorId, "NIKON   ", 8) == 0
            && memcmp(inquiry.ProductId, "NIKON DSC E", 11) == 0 )
         {
            return new USB_DSC( hDSC, DSC_DRIVE );
         }

         CloseHandle( hDSC );
      }
   }
   return NULL;
}

bool CheckUSB(char camera_drive, char* buf, unsigned size)
{
   int len = wsprintf( buf, "Open drive %c: ", camera_drive );
   buf += len;
   size -= len;

   DSC_DRIVE[4] = camera_drive;

   SetLastError( NOERROR );

   HANDLE hDSC = CreateFile(DSC_DRIVE, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
      NULL, OPEN_EXISTING, 0, NULL );

   if( INVALID_HANDLE_VALUE == hDSC )
   {
      len = wsprintf( buf, "\r\nERROR: " );
      buf += len;
      size -= len;

      len = FormatMessage( FORMAT_MESSAGE_MAX_WIDTH_MASK	|
         FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
         NULL, GetLastError(), 0, buf, size-2, NULL );

      buf += len;

      strcpy( buf, "\r\n" );
      return FALSE;
   }
   else
   {
      len = wsprintf( buf, "SUCCESS \r\n" );
      buf += len;
      size -= len;
   }

   len = wsprintf( buf, "Inquiry drive:\r\n" );
   buf += len;
   size -= len;

   TInquiryData inquiry;

   SetLastError( NOERROR );

   bool ret = GetDeviceName(hDSC, inquiry);

   if( ret )
   {
      wsprintf( buf, "%sDevice: %.8s  %.16s %.4s\r\n",
         inquiry.RemovableMedia ? "Removable ":"",
         inquiry.VendorId, inquiry.ProductId, inquiry.ProductRevisionLevel);
   }
   else
   {
      len = wsprintf( buf, "ERROR: " );
      buf += len;
      size -= len;

      len = FormatMessage( FORMAT_MESSAGE_MAX_WIDTH_MASK	|
         FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
         NULL, GetLastError(), 0, buf, size-2, NULL );

      buf += len;

      strcpy( buf, "\r\n" );
   }

   CloseHandle( hDSC );
   return TRUE;
}

